% AC.lhs
%
% This file is part of CalcCheck
%
% Copyright (c) 2011  Wolfram Kahl
% All rights reserved
%
\section{|AC|: General-Purpose AC-Matching}

We follow \cite{Contejean-2004} for AC matching.
\edcomm{WK}{Only-C operators not yet implemented.}

\edcomm{WK}{Variable binders need to be added! Substitution?}

%{{{ module AC
\begin{code}
{-# LANGUAGE ScopedTypeVariables, PatternGuards, FlexibleContexts #-}
module AC
--  ( Term(), var, apply, applyAC, fold )
  where

import Utils.Q
import Utils.List

import Data.Map (Map)
import qualified Data.Map as Subst

import Data.Set (Set)
import qualified Data.Set as Set

import Data.List (nubBy, groupBy, sortBy)
import Data.Function (on)

import Control.Arrow (second)
import Control.Monad (MonadPlus(..))
import Control.Monad.State (MonadState, get, put, evalStateT)
\end{code}
%}}}

\subsection{Canonical forms of terms: |data Term|}

%{{{ Canonical forms of terms: data Term
Canonical forms of terms ---
for the total term ordering, we assume total orders on
variables and operators:
\begin{code}
data Term var op cnst
  = Const cnst
  | Appl op [Term var op cnst]  -- non-AC operators
  | AC op [Term var op cnst]    -- at least two, and sorted
  | C op (Term var op cnst) (Term var op cnst)  -- sorted
  | Var var
  deriving (Eq, Ord)
\end{code}
%}}}

%{{{ varInTerm
\begin{code}
varInTerm :: Eq var => var -> Term var op cnst -> Bool
varInTerm v (Var x) = v == x
varInTerm _ (Const _) = False
varInTerm v (Appl _ ts) = any (varInTerm v) ts
varInTerm v (AC _ ts) = any (varInTerm v) ts
varInTerm v (C _ t1 t2) = varInTerm v t1 || varInTerm v t2
\end{code}
%}}}

%{{{ varEqTerm
\begin{code}
varEqTerm :: Eq var => var -> Term var op cnst -> Bool
varEqTerm v (Var v') = v == v'
varEqTerm _ _ = False
\end{code}
%}}}

%{{{ Abstract interface to |Term|:
Abstract interface to |Term|:

\edcomm{WK}{Unused!
\begin{spec}
fold  :: (var -> r)
      -> (cnst -> r)
      -> (op -> [r] -> r)
      -> (op -> r -> r -> r)
      -> Term var op cnst
      -> r
fold v _ _ap  _ac  (Var x) = v x
fold _ k _ap  _ac  (Const c) = k c
fold v k ap   ac   (Appl op ts) = ap op (map (fold v k ap ac) ts)
fold v k ap   ac   (AC op ts) = foldr1 (ac op) (map (fold v k ap ac) ts)
fold v k ap   ac   (C op t1 t2) = ac op (fold v k ap ac t1) (fold v k ap ac t2)

fold'  :: (var -> r)
       -> (cnst -> r)
       -> (op -> [r] -> r)
       -> (op -> [r] -> r)
       -> Term var op cnst
       -> r
fold' v _ _ap  _ac  (Var x) = v x
fold' _ k _ap  _ac  (Const c) = k c
fold' v k ap   ac   (Appl op ts)  = ap op (map (fold' v k ap ac) ts)
fold' v k ap   ac   (AC op ts)    = ac op (map (fold' v k ap ac) ts)
\end{spec}
}
\begin{code}
var :: var -> Term var op cnst
var = Var

unVar :: Term var op cnst -> Maybe var
unVar (Var v) = Just v
unVar _ = Nothing

apply :: op -> [Term var op cnst] -> Term var op cnst
apply = Appl

applyAC :: (Ord var, Ord op, Ord cnst) => op -> Term var op cnst -> Term var op cnst -> Term var op cnst
applyAC op t u = let
  opArgs x = case x of
    AC op1 ts | op == op1  -> ts
    _                      -> [x]
  in AC op (merge (opArgs t) (opArgs u))

applyACs :: (Ord var, Ord op, Ord cnst) => op -> [Term var op cnst] -> Term var op cnst
applyACs op = foldr1 (applyAC op)

applyC :: (Ord var, Ord op, Ord cnst) => op -> Term var op cnst -> Term var op cnst -> Term var op cnst
applyC op t1 t2 = if t1 <= t2 then C op t1 t2 else C op t2 t1

applyCs :: (Ord var, Ord op, Ord cnst) => op -> [Term var op cnst] -> Term var op cnst
applyCs op [t1, t2] = applyC op t1 t2
applyCs _ ts = error $ "Internal error: applyCs called with " ++ shows (length ts) " arguments"
\end{code}
%}}}

%{{{ mkAC
Internal re-construction function --- non-empty sorted list expected:
\begin{code}
mkAC :: op -> [Term var op cnst] -> Term var op cnst
mkAC _op [] = error "``Impossible'': mkAC []"
mkAC _op [t] = t
mkAC op ts = AC op ts
\end{code}
%}}}

\subsection{|ShowS| Functions}

%{{{ showsTerm
\begin{code}
showsTerm :: (Show var, Show op, Show cnst) => Term var op cnst -> ShowS

showsTerm = h
  where
    parens f = ('(' :) . f . (')' :)
    space = (' ' :)
    spaced f = space . f . space
    h (Var v) = shows v
    h (Const c) = shows c
    h (Appl op []) = shows op
    h (Appl op [t]) = shows op . h t
    h (Appl op [t,u]) = parens $ h t . spaced (shows op) . h u
    h (Appl op ts) = shows op . parens
        (foldr1 (\ l r -> l . (", " ++) . r) $ map h ts)
    h (AC op ts) = parens
      (foldr1 (\ l r -> l . spaced (shows op) . r) $ map h ts)
    h (C op t1 t2) = parens
      (h t1 . spaced (shows op) . h t2)
\end{code}
%}}}

%{{{ showsPrecTerm
\begin{code}
showsPrecTerm :: (Ord prec)
  => (var -> ShowS)             -- |showsVar|
  -> (cnst -> ShowS)            -- |showsConst|
  -> (op -> [ShowS] -> ShowS)   -- |showsOp|
  -> (op -> ShowS)              -- |infixOp|
  -> (op -> op -> Bool)         -- |mutual|
  -> (op -> prec)               -- |opPrec|
  -> Maybe op                   -- ||
  -> prec
  -> Term var op cnst
  -> ShowS

showsPrecTerm showsVar showsConst showsOp infixOp mutual opPrec = h
  where
    parens False s = s
    parens True s = ('(' :) . s . (')' :)
    needParens mOp p op p' = case compare p p' of
      GT -> True
      EQ -> not (maybe False (mutual op) mOp)
      LT -> False
    h _    _ (Var v) = showsVar v
    h _    _ (Const c) = showsConst c
    h _    _ (Appl op []) = showsOp op []
    h mOp  p (Appl op ts) = let p' = opPrec op in parens
      (needParens mOp p op p')
      (showsOp op $ map (h (Just op) p') ts)
    h mOp  p (AC op ts) = let p' = opPrec op in parens
      (needParens mOp p op p')
      (foldr1 (\ l r -> l . infixOp op . r) $ map (h (Just op) p') ts)
    h mOp  p (C op t1 t2) = let p' = opPrec op in parens
      (needParens mOp p op p')
      (h (Just op) p' t1 . infixOp op . h (Just op) p' t2)
\end{code}
%}}}

\subsection{AC Matching}

%{{{ datatypes
Partial solutions:

\begin{code}
data PSol var op cnst = PSol op var (Term var op cnst)
\end{code}

\begin{code}
type Subst var op cnst = Map var (Term var op cnst)
\end{code}

\begin{code}
type NotOccurs var op cnst = ([var], Term var op cnst)
\end{code}

\begin{code}
type TermPair var op cnst = (Term var op cnst, Term var op cnst)
\end{code}

Simple matching problem:
\begin{code}
data SMP var op cnst = SMP
  { unsolved      :: [TermPair var op cnst]     -- can be considered as set
  , partlySolved  :: [(var, PSol var op cnst)]  -- to be considered as stack
  , solved        :: Subst var op cnst          -- no reason to consider as list (see end of 3.1 in \cite{Contejean-2004})
  }
\end{code}
%}}}

%{{{ initSMP
\textbf{Init}:
\begin{code}
initSMP :: [TermPair var op cnst] -> SMP var op cnst
initSMP ps = SMP ps [] Subst.empty
\end{code}
%}}}

%{{{ extract
\begin{code}
extract  :: (Ord var, Ord op, Ord cnst, Show var, Show op, Show cnst)
         => [(var, PSol var op cnst)]
         -> Subst var op cnst
         -> Subst var op cnst
extract [] su = su
extract ((x, PSol op y s) : part) su = case Subst.lookup y su of
  Nothing -> error $ "extract " ++ shows x " = " ++ shows y " " ++ shows op " " ++ showsTerm s ": " ++ shows y " not found!"
  Just t -> extract part $ Subst.insert x (applyAC op t s) $ Subst.delete y su
\end{code}
%}}}

%{{{ match :: [var] -> [TermPair var op cnst] -> [Subst var op cnst]
\begin{code}
-- |match  :: (MonadPlus m, Show var, Ord var, Ord op)|  -- principal type
       -- |=> [var] -> [TermPair var op cnst] -> m (Subst var op cnst)|
match  :: (Show var, Show op, Show cnst, Ord var, Ord op, Ord cnst)
       => [var] -> [TermPair var op cnst] -> [Subst var op cnst]
match freshVars ps = let
    fresh = mkFresh "AC.match"
  in evalStateT (match1 fresh ps) freshVars

match1 :: (Ord op, Show op, Ord var, Show var, Ord cnst, Show cnst, MonadPlus m)
  => m var -> [TermPair var op cnst] -> m (Subst var op cnst)
match1 fresh ps = match0 fresh $ initSMP ps

match0 :: (Ord op, Show op, Ord var, Show var, Ord cnst, Show cnst, MonadPlus m)
  => m var -> SMP var op cnst -> m (Subst var op cnst)
match0 fresh = run
  where
    run (SMP [] part su) = return $ extract part su
    run smp = step fresh smp >>= run

mkFresh :: (MonadState [var] m) => String -> m var
mkFresh caller = do
      vs <- get
      case vs of
        [] -> error $ caller ++ ": out of freshVars!"
        (v : vs') -> do
           put vs'
           return v
\end{code}
%}}}

%{{{ step :: m var -> SMP var op cnst -> m (SMP var op cnst)
\begin{code}
step :: (MonadPlus m, Ord var, Ord op, Ord cnst)
  => m var
  -> SMP var op cnst
  -> m (SMP var op cnst)
step _fresh (SMP ((Var x, s) : u) part su)
  | Just s' <- Subst.lookup x su
  = if s' == s  then return $ SMP u part su            -- \textbf{Merge}${}_S$
                else fail "Clash_S"                    -- \textbf{Clash}${}_S$
  | Just (PSol op' y s') <- lookup x part
  = case s of
      AC op ss | op == op', (ss', _ : ss'') <- span (s' /=) ss
        -> return $ SMP ((Var y, mkAC op (ss' ++ ss'')) : u) part su  -- \textbf{Merge}${}_P$
      _ -> fail "Clash_P_AC"                         -- \textbf{Clash}${}_P$
  | otherwise
  = return $ SMP u part (Subst.insert x s su)        -- \textbf{Solve}

step _fresh (SMP ((Const c, s) : u) part su) = case s of
  Const d | c == d  -> return $ SMP u part su   -- \textbf{Dec}
  _                 -> fail "Clash"             -- \textbf{Clash}

step _fresh (SMP ((Appl f ps, s) : u) part su) = case s of
  Appl g ss | f == g  -> return $ SMP (zip ps ss ++ u) part su   -- \textbf{Dec}
  _                   -> fail "Clash"                            -- \textbf{Clash}

step _fresh (SMP ((C op p1 p2, s) : u) part su) = case s of
  C op' s1 s2 | op == op'  ->           return (SMP ((p1, s1) : (p2, s2) : u) part su)
                               `mplus`  return (SMP ((p1, s2) : (p2, s1) : u) part su) -- \textbf{Dec$_C$}
  _                        ->  fail "Clash"                            -- \textbf{Clash}

step fresh (SMP ((AC op ps@(p : ps'), AC op' ss) : u) part su) | op == op', m <= n = case p of
  Var x -> case Subst.lookup x su of
    Just s -> case s of
      AC op'' ss'  | op'' == op, pred m + length ss' <= n
                   , Just sd <- sortedListDiff ss ss'
        -> return $ SMP ((mkAC op ps', mkAC op' sd) : u) part su -- \textbf{AC}$_{S_=}$
      _ -> case find s ss of
        Just ss' -> return $ SMP ((mkAC op ps', mkAC op' ss') : u) part su -- \textbf{AC}$_{S_{\neq}}$
        _ -> mzero
    Nothing -> case lookup x part of
      Just (PSol op'' y t) -> if op == op''
        then if shorter
          then case find t ss of
            Just ss' -> return $ SMP ((mkAC op $ Var y : ps', mkAC op' ss') : u) part su -- \textbf{AC}$_{P_{=}}$
            Nothing -> mzero
          else mzero
        else do  -- \textbf{AC}$_{P_{\neq}}$
          (ts, ss') <- findInAC op'' t ss
          return $ SMP ((Var y, mkAC op'' ts) : (mkAC op ps', mkAC op' ss') : u) part su
      Nothing -> do
         (s, ss') <- foldr (mplus . return) mzero $ uninsert ss
         return $ SMP ((mkAC op ps', mkAC op' ss') : u) part (Subst.insert x s su) -- \textbf{AC}$_{\neq}$
       `mplus`
          if shorter
          then do -- \textbf{AC}$_{=}$
            y <- fresh
            (s, ss') <- foldr (mplus . return) mzero $ uninsert ss
            return $ SMP ((mkAC op $ Var y : ps', mkAC op' ss') : u) ((x, PSol op' y s) : part) su
          else mzero
  _ -> do                                             -- \textbf{AC}
    (q, ss') <- findOp p ss
    return $ SMP (q : (mkAC op ps', mkAC op' ss') : u) part su
  where
    m = length ps
    n = length ss
    shorter = m < n

step _ _ = fail "Fail" --- \textbf{Fail} or \textbf{Clash}
\end{code}
%}}}

\subsection{Utility Functions}

%{{{ findOp
For \textbf{AC} \edcomm{WK}{convert to returning a list, and add |nub|?}:
\begin{code}
findOp :: (MonadPlus m, Eq op, Eq cnst)
  => Term var op cnst
  -> [Term var op cnst]
  -> m ((TermPair var op cnst), [Term var op cnst])
findOp p = h id
  where
    matchS s f ss = if headEqual p s then mplus $ return ((p, s), f ss) else id
    h _ [] = mzero
    h f (s : ss) = matchS s f ss $ h (f . (s :)) ss
\end{code}

\begin{code}
headEqual :: (Eq op, Eq cnst) => Term var op cnst -> Term var op cnst -> Bool
headEqual (Const c)    (Const c')    = c == c'
headEqual (Appl op _)  (Appl op' _)  = op == op'
headEqual (AC op _)    (AC op' _)    = op == op'
headEqual _ _ = False
\end{code}
%}}}

%{{{ findInAC
For \textbf{AC}$_{P_{\neq}}$:
\begin{code}
findInAC :: (MonadPlus m, Eq var, Eq op, Eq cnst)
  => op
  -> Term var op cnst
  -> [Term var op cnst]
  -> m ([Term var op cnst], [Term var op cnst])
findInAC op t = h id
  where
    h _ [] = mzero
    h f (s : ss) = case s of
      AC op' ts | op == op' -> let
          rest = h (f . (s :)) ss
        in case find t ts of
          Just ts'  -> return (ts', f ss) `mplus` rest
          Nothing   -> rest
      _ -> mzero          
\end{code}
%}}}

%{{{ \subsection{Equation Utilities}
\subsection{Equation Utilities}

\begin{code}
termSize :: Term var op cnst -> Int
termSize (Var _) = 1
termSize (Const _) = 1
termSize (Appl _op ts) = 2 + length ts
termSize (C _op _t1 _t2) = 4
termSize (AC _op ts) = 2 + length ts
\end{code}

\begin{code}
compareBySize :: (Ord var, Ord op, Ord cnst) => Term var op cnst -> Term var op cnst -> Ordering
compareBySize t1 t2 = compare (termSize t2, t2) (termSize t1, t2)
\end{code}

\begin{code}
compareTermPair :: (Ord var, Ord op, Ord cnst) => TermPair var op cnst -> TermPair var op cnst -> Ordering
compareTermPair (t1,u1) (t2,u2) = case compareBySize t1 t2 of
  EQ -> compareBySize u2 u1
  x -> x
\end{code}

%{{{ substIsSubid
\begin{code}
substIsSubid :: (Eq var) => Subst var op cnst -> Bool
substIsSubid = Subst.foldrWithKey (\ v t r -> varEqTerm v t && r) True
\end{code}
%}}}

%{{{ applySubst
\begin{code}
applySubst :: (Ord var, Ord op, Ord cnst) => Subst var op cnst -> Term var op cnst -> Term var op cnst
applySubst su = qDo q
  where
    q t@(Var v) = case Subst.lookup v su of
      Just t' | t == t' -> Nothing
      m -> m
    q (Const _) = Nothing
    q (Appl op ts) = fmap (apply op) $ qList q ts
    q (C op t1 t2) = q2 (applyC op) q q t1 t2
    q (AC op ts) = fmap (foldr1 (applyAC op)) $ qList q ts
\end{code}
%}}}

\begin{code}
allVars :: (var -> Bool) -> Term var op cnst -> Bool
allVars p (Var v) = p v
allVars _ (Const _) = True
allVars p (Appl _op ts) = all (allVars p) ts
allVars p (C _op t1 t2) = allVars p t1 && allVars p t2
allVars p (AC _op ts) = all (allVars p) ts
\end{code}

\begin{code}
vars :: (Ord var) => Term var op cnst -> Set var
vars (Var v) = Set.singleton v
vars (Const _) = Set.empty
vars (Appl _op ts) = Set.unions $ map vars ts
vars (C _op t1 t2) = Set.union (vars t1) (vars t2)
vars (AC _op ts) = Set.unions $ map vars ts
\end{code}

\begin{code}
noNewVars :: (Ord var) => TermPair var op cnst -> Bool
noNewVars (lhs, rhs) = allVars (`Set.member` vars lhs) rhs
\end{code}

\begin{code}
substInverse :: (Ord var, Ord op) => Subst var op cnst -> Maybe (Subst var op cnst)
substInverse su = let
    h k (Var v) m@(Just su') = case Subst.lookup v su' of
      Just (Var k') -> if k == k' then m else Nothing
      Just _ -> error "substInverse: Impossible"
      Nothing -> Just $ Subst.insert v (Var k) su'
    h _ _ _ = Nothing
  in Subst.foldrWithKey h (Just Subst.empty) su
\end{code}

\begin{code}
matchIso :: (Ord var, Ord op, Ord cnst, Show var, Show op, Show cnst)
  => [var] -> [TermPair var op cnst] -> Maybe (Subst var op cnst, Subst var op cnst)
matchIso freshVars tps = case match freshVars tps of
  [] -> Nothing
  su : _ -> case substInverse su of
    Just su' | all (\ (t,u) -> applySubst su' u == t) tps  ->  Just (su, su')
    _                                                      ->  Nothing
\end{code}

\begin{code}
equEquiv :: (Ord var, Ord op, Ord cnst, Show var, Show op, Show cnst)
  => [var] -> TermPair var op cnst -> TermPair var op cnst -> Bool
equEquiv freshVars (t1,t2) (u1,u2) = case matchIso freshVars [(t1,u1),(t2,u2)] of
  Nothing -> False
  Just _ -> True
\end{code}

\begin{code}
nubEqus :: (Ord var, Ord op, Ord cnst, Show var, Show op, Show cnst)
  => [var] -> [TermPair var op cnst] -> [TermPair var op cnst]
nubEqus freshVars = concatMap (nubBy $ equEquiv freshVars) .
   groupBy (\ (l1,_) (l2,_) -> termSize l1 == termSize l2) .
   sortBy compareTermPair . filter noNewVars
\end{code}

Since AC-matching of left-hand sides takes a lot of effort,
we share it by organising rules into groups with the same left-hand side.

\begin{code}
type RuleGroup var op cnst r = (Term var op cnst, [(Term var op cnst, r)])
\end{code}

When grouping equations according to isomorphic LHSs,
we first pre-group according to term size to avoid superfluous |match| calls.

\begin{code}
groupEqus :: forall var op cnst r . (Ord var, Ord op, Ord cnst, Show var, Show op, Show cnst)
  => [var] -> [(TermPair var op cnst, r)] -> [RuleGroup var op cnst r]
groupEqus freshVars tps = let
    group  :: [(TermPair var op cnst, r)] -> [[(TermPair var op cnst, r)]]
    group  = map (map snd) . groupBy ((==) `on` fst)
           . sortBy (compare `on` second fst)
           . map (\ e -> (termSize (fst $ fst e), e))
    sameLHS ((t1,_u1),_r1) ((t2,u2),r2) = case matchIso freshVars [(t1,t2)] of
      Just (_, su) -> Just (applySubst su u2, r2)
      Nothing -> Nothing
    bundle [] = []
    bundle (e@((lhs,rhs),r) : es) = case partitionM (sameLHS e) es of
      (rhss', es') -> (lhs, nubBy ((==) `on` fst) $ (rhs, r) : rhss') : bundle es'
  in concatMap bundle $ group tps
\end{code}
%}}}

\subsection{Special Rules}

Symmetry of equivalence $p \equiv q \equiv q \equiv p$
is typically used to eliminate double occurrences
of terms. Since this leads to combinatorial explosion with AC matching,
we provide a special function.
|removeSelfInverse| removes at least one self-inverse situation,
while |removeSelfInverse0| removes zero or more.

\begin{code}
removeSelfInverse0, removeSelfInverse :: (Eq var, Eq op, Eq cnst) => (op -> Bool) -> Term var op cnst -> [Term var op cnst]
removeSelfInverse0 opPred (AC op ts)
  | opPred op = map (mkAC op) $ ts : filter (not . null) (sortedRemDupls ts)
  | otherwise = map (mkAC op) $ mapM (removeSelfInverse0 opPred) ts
removeSelfInverse0 opPred (Appl op ts) = map (Appl op) $ mapM (removeSelfInverse0 opPred) ts
removeSelfInverse0 _ t = [t]

removeSelfInverse opPred (AC op ts)
  | opPred op = map (mkAC op) $ filter (not . null) $ sortedRemDupls ts
  | otherwise = map (mkAC op) $ removeSelfInverses opPred ts
removeSelfInverse opPred (Appl op ts) = map (Appl op) $ removeSelfInverses opPred ts
removeSelfInverse _ _ = []

removeSelfInverses :: (Eq var, Eq op, Eq cnst) => (op -> Bool) -> [Term var op cnst] -> [[Term var op cnst]]
removeSelfInverses _opPred [] = []
removeSelfInverses opPred (t : ts) = case removeSelfInverse opPred t of
  [] -> map (t :) $ removeSelfInverses opPred ts
  us -> [u : vs | u <- us , vs <- mapM (removeSelfInverse0 opPred) ts]
\end{code}


%{{{ EMACS lv
% Local Variables:
% folded-file: t
% fold-internal-margins: 0
% eval: (fold-set-marks "%{{{ " "%}}}")
% eval: (fold-whole-buffer)
% end:
%}}}
