% Utils/List.lhs
%
% This file is part of CalcCheck
%
% Copyright (c) 2011--2013  Wolfram Kahl
% All rights reserved
%
\section{|Utils.List|: Useful List Functions}

\begin{code}
module Utils.List where

import qualified Data.Set as Set

import Data.List (groupBy)
import Data.Function (on)
\end{code}

%{{{ sortedListDiff :: Ord a => [a] -> [a] -> Maybe [a]
|sortedListDiff xs ys = Just zs| $\equiv$ |merge' ys zs = xs|
where |merge'| does not eliminate duplicates.
\begin{code}
sortedListDiff :: Ord a => [a] -> [a] -> Maybe [a]
sortedListDiff xs [] = Just xs
sortedListDiff [] _ys = Nothing
sortedListDiff (x : xs') ys@(y : ys') = case x `compare` y of
  LT -> fmap (x :) $ sortedListDiff xs' ys
  EQ -> sortedListDiff xs' ys'
  GT -> Nothing
\end{code}
%}}}

%{{{ find :: Eq a => a -> [a] -> Maybe [a]
|find| is ``remove if found'':
\begin{spec}
find x xs = case span (x /=) xs of
  (ys, _ : zs)  -> Just (ys ++ zs)
  _             -> Nothing
\end{spec}

\begin{code}
find :: Eq a => a -> [a] -> Maybe [a]
find x = h id
  where
    h _ [] = Nothing
    h f (y : ys) = if x == y then Just (f ys) else h (f . (y :)) ys
\end{code}
%}}}

%{{{ uninsert :: [a] -> [(a, [a])]
This is designed for ordered argument lists
where equality |(==)| is sufficiently strong
so that we can avoid generating duplicates based on it.
\begin{code}
uninsert :: Eq a => [a] -> [(a, [a])]
uninsert [] = []
uninsert (u : us) = (u, us) : h u (u:) us
  where
    h _ _ [] = []
    h z f (x : xs) = (if x == z then id else ((x, f xs) :)) $ h x (f . (x :)) xs
\end{code}
%}}}

%{{{ nubOrd :: Ord a => [a] -> [a]
\begin{code}
nubOrd :: Ord a => [a] -> [a]
nubOrd = h Set.empty
  where
    h _ [] = []
    h s (x : xs) =  if x `Set.member` s
                    then h s xs
                    else x : h (Set.insert x s) xs
\end{code}
%}}}

%{{{ nubOrdOn :: Ord a => (b -> a) -> [b] -> [b]
\begin{code}
nubOrdOn :: Ord a => (b -> a) -> [b] -> [b]
nubOrdOn f = h Set.empty
  where
    h _ [] = []
    h s (x : xs) =  let y = f x in if y `Set.member` s
                    then h s xs
                    else x : h (Set.insert y s) xs
\end{code}
%}}}

%{{{ merge :: Ord a => [a] -> [a] -> [a]
\begin{code}
merge :: Ord a => [a] -> [a] -> [a]
merge [] ys = ys
merge xs@(x : xs') ys = merge' xs x xs' ys
  where
    merge' us _u _us' [] = us
    merge' us u us' vs@(v : vs') = if u < v
      then u : merge' vs v vs' us'
      else v : merge' us u us' vs'
\end{code}
%}}}

%{{{ lookup' :: Eq a => a -> [(a,b)] -> Maybe ([(a,b)], b)
\begin{code}
lookup' :: Eq a => a -> [(a,b)] -> Maybe ([(a,b)], b)
lookup' x = h id
  where
    h _ [] = Nothing
    h f (p@(y,v) : ps) = if x == y then Just (f ps, v) else h (f . (p :)) ps
\end{code}
%}}}

%{{{ groupFst, groupSnd
\begin{code}
groupFst :: Eq a => [(a,b)] -> [(a,[b])]
groupFst = map h . groupBy ((==) `on` fst)
  where h ps@(~((x,_) : _)) = (x, map snd ps)
\end{code}
%}}}

%{{{ splitEithers
\begin{code}
splitEithers :: [Either a b] -> ([a], [b])
splitEithers = h id id
  where
    h l r [] = (l [], r [])
    h l r (Left a : es) = h (l . (a :)) r es
    h l r (Right b : es) = h l (r . (b :)) es
\end{code}
%}}}

%{{{ partitionM :: (a -> Maybe b) -> [a] -> ([b], [a])
\begin{code}
partitionM :: (a -> Maybe b) -> [a] -> ([b], [a])
partitionM f = h id id
  where
    h mkB mkA [] = (mkB [], mkA [])
    h mkB mkA (x : xs) = case f x of
       Nothing  -> h mkB (mkA . (x :)) xs
       Just y   -> h (mkB . (y :)) mkA xs
\end{code}
%}}}

%{{{ sortedRemDupls :: (Eq a) => [a] -> [[a]]
|sortedRemDupls xs| returns all different sublists of the sorted list |xs|
that can be obtained by eliminating at least one pair of duplicates.

\begin{code}
sortedRemDupls :: (Eq a) => [a] -> [[a]]
sortedRemDupls = init .
   foldr (\ fs rs -> foldr (\ r ss -> foldr (\ f t -> f r : t) ss fs) [] rs) [[]] . h
  where
    h [] = [[id]]
    h (x : xs@(y : ys))  | x == y     =  [id, (x :) . (y :)] : h ys
                         | otherwise  =  [(x :)] : h xs
    h [x] = [[(x :)]]
\end{code}
%}}}

%{{{ mapHead :: (a -> a) -> [a] -> [a]
\begin{code}
mapHead :: (a -> a) -> [a] -> [a]
mapHead _ [] = []
mapHead f (x : xs) = f x : xs
\end{code}
%}}}

%{{{ EMACS lv
% Local Variables:
% folded-file: t
% fold-internal-margins: 0
% eval: (fold-set-marks "%{{{ " "%}}}")
% eval: (fold-whole-buffer)
% end:
%}}}
